import logging

from flask import Blueprint, request, json
from flask_jwt_extended import jwt_required
from flask_restful import Api, Resource

import portal.service.auth
import portal.service.general
import portal.service.users
from portal.database.models import User
from portal.rest import rest_helpers
from portal.rest.schemas import user_schema, user_schema_reduced, users_schema, \
    password_change_schema, submissions_schema, reviews_schema, roles_schema, courses_schema, \
    groups_schema
from portal.service.errors import PortalAPIError, ForbiddenError
from portal.service import permissions, general
from portal.service.auth import find_client
from portal.service.users import find_users_filtered, find_groups_filtered, find_roles_filtered, \
    find_reviews, find_submissions_filtered, create_user, update_user, delete_user, \
    update_user_password
from portal.service.general import find_user
from portal.tools.decorators import error_handler

users_blueprint = Blueprint('users', __name__)
users_api = Api(users_blueprint)

log = logging.getLogger(__name__)


class UserResource(Resource):
    @error_handler
    @jwt_required
    def get(self, uid):
        client = find_client()
        user = find_user(uid)
        # authorization
        if permissions.check_component(component=client) \
                or permissions.check_sysadmin(client) \
                or client == user:
            return user_schema.dump(user)[0], 200
        return user_schema_reduced.dump(user)[0], 200

    @error_handler
    @jwt_required
    def put(self, uid):
        client = find_client()
        user = find_user(uid)
        # authorization
        if not (permissions.check_component(component=client)
                or permissions.check_sysadmin(client)
                or client == user):
            raise ForbiddenError(uid=client.id)
        full = False
        if permissions.check_sysadmin(client):
            full = True

        data = rest_helpers.parse_request_data(
            user_schema, action='update', resource='user'
        )
        update_user(user, data, full)
        return '', 204

    @error_handler
    @jwt_required
    def delete(self, uid):
        client = find_client()
        # authorization:
        if not permissions.check_sysadmin(client):
            raise ForbiddenError(client.username)

        user = find_user(uid)
        delete_user(user)
        return '', 204


class UserPassword(Resource):
    @error_handler
    @jwt_required
    def put(self, uid):
        client = portal.service.auth.find_client()
        user = portal.service.general.find_user(uid)
        # authorization
        if not (permissions.check_component(component=client)
                or permissions.check_sysadmin(client)
                or client == user):
            raise ForbiddenError(uid=client.id)

        data = rest_helpers.parse_request_data(
            password_change_schema, action='update', resource='user_password'
        )
        update_user_password(client, data, user)
        return '', 204


class UserList(Resource):
    @error_handler
    @jwt_required
    def get(self):
        client = portal.service.auth.find_client()
        # authorization
        if not permissions.check_sysadmin(client):
            raise ForbiddenError(uid=client.id)

        course_id = request.args.get('course')
        group_id = request.args.get('group')
        if group_id and not course_id:
            raise PortalAPIError(400,
                                 "Invalid filter combination at /get/users: "
                                 "missing course id.")

        filtered_users = find_users_filtered(course_id=course_id, group_id=group_id)
        return users_schema.dump(filtered_users)[0], 200

    @error_handler
    @jwt_required
    def post(self):
        client = portal.service.auth.find_client()
        # authorization
        if not permissions.check_sysadmin(client):
            raise ForbiddenError(uid=client.id)

        data = rest_helpers.parse_request_data(user_schema, action='create', resource='user')
        new_user = create_user(**data)
        return user_schema.dump(new_user)[0], 201


class UserSubmissionList(Resource):
    @staticmethod
    def find_groups_intersection(user1: User, user2: User):
        temp = set(user2.groups)
        res = [value for value in user1.groups if value in temp]
        return res

    @error_handler
    @jwt_required
    def get(self, uid):
        client = portal.service.auth.find_client()
        user = portal.service.general.find_user(uid)

        course_id = request.args.get('course')
        project_ids = request.args.getlist('project')
        course = general.find_course(course_id, throws=False)
        if project_ids and not course_id:
            raise PortalAPIError(400,
                                 f"Invalid filter combination at "
                                 f"/get/users/{uid}/submissions: missing course id.")
        # authorization
        if course_id is not None and course_id != '':
            if permissions.check_component(component=client) \
                    or permissions.check_sysadmin(client) \
                    or permissions.check_user(client, course, ['read_submissions_all']) \
                    or (permissions.check_user(client, course, ['read_submissions_groups']
                        and self.find_groups_intersection(client, user))) \
                    or (permissions.check_user(client, course, ['read_submissions_own']) and client == user):
                submissions = find_submissions_filtered(user=user, course_id=course_id,
                                                        project_ids=project_ids)
                return submissions_schema.dump(submissions)[0], 200
        elif (course_id is None or course_id == '') and client.is_admin:
            return submissions_schema.dump(user.submissions)[0], 200
        elif (course_id is None or course_id == '') and client == user:
            # rare use case (owner does not have read access to all their submissions) - not filtered now
            return submissions_schema.dump(user.submissions)[0], 200

        raise ForbiddenError(uid=client.id)


class UserRoleList(Resource):
    @error_handler
    @jwt_required
    def get(self, uid):
        client = portal.service.auth.find_client()
        user = portal.service.general.find_user(uid)
        # authorization
        if not (permissions.check_component(component=client)
                or permissions.check_sysadmin(client)
                or client == user):
            raise ForbiddenError(uid=client.id)

        course_id = request.args.get('course')

        roles = find_roles_filtered(user=user, course_id=course_id)
        return roles_schema.dump(roles)[0], 200


class UserCourseList(Resource):
    @error_handler
    @jwt_required
    def get(self, uid):
        client = portal.service.auth.find_client()
        user = portal.service.general.find_user(uid)
        # authorization
        if not (permissions.check_component(component=client)
                or permissions.check_sysadmin(client)
                or client == user):
            raise ForbiddenError(uid=client.id)

        return courses_schema.dump(user.courses)[0], 200


class UserGroupList(Resource):
    @error_handler
    @jwt_required
    def get(self, uid):
        client = portal.service.auth.find_client()
        user = portal.service.general.find_user(uid)
        # authorization
        if not (permissions.check_sysadmin(client) or client == user):
            raise ForbiddenError(uid=client.id)
        course_id = request.args.get('course')

        groups = find_groups_filtered(user=user, course_id=course_id)
        return groups_schema.dump(groups)[0], 200


class UserReviewList(Resource):
    @error_handler
    @jwt_required
    def get(self, uid):
        client = portal.service.auth.find_client()
        user = portal.service.general.find_user(uid)
        # authorization
        if not (permissions.check_component(component=client)
                or permissions.check_sysadmin(client)
                or client == user):
            raise ForbiddenError(uid=client.id)

        reviews = find_reviews(user)
        return reviews_schema.dump(reviews)[0], 200


class UserEffectivePermissions(Resource):
    @error_handler
    @jwt_required
    def get(self, uid):
        client = portal.service.auth.find_client()
        user = portal.service.general.find_user(uid)
        # authorization
        if not (permissions.check_component(component=client)
                or permissions.check_sysadmin(user)
                or client == user):
            raise ForbiddenError(uid=client.id)

        course_id = request.args.get('course')
        perm = permissions.get_effective_permissions(user, course_id)
        return perm, 200


users_api.add_resource(UserResource, '/<string:uid>')
users_api.add_resource(UserPassword, '/<string:uid>/password')
users_api.add_resource(UserList, '')
users_api.add_resource(UserSubmissionList, '/<string:uid>/submissions')
users_api.add_resource(UserRoleList, '/<string:uid>/roles')
users_api.add_resource(UserCourseList, '/<string:uid>/courses')
users_api.add_resource(UserGroupList, '/<string:uid>/groups')
users_api.add_resource(UserReviewList, '/<string:uid>/reviews')
users_api.add_resource(UserEffectivePermissions, '/<string:uid>/permissions')
